// Image.cpp: implementation of the CImage class.
// Author : Lamon Pierre 2000
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "USBCam.h"
#include "Image.h"
#include "Utils.h"
#include "Wingdi.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CImage::CImage()
{

}

CImage::CImage(long width, long height, short bPerPix)
{
	DWORD dwBitsSize;
	DWORD dwBitsPix;
	dwBitsPix = bPerPix * width * height;   // size in bytes of the image
	dwBitsSize = dwBitsPix;
	dwBitsSize += sizeof (BITMAPFILEHEADER); // Add size of file structure
	dwBitsSize += sizeof (BITMAPINFOHEADER); // Add size of info structure
	
    hDIB = (HDIB) ::GlobalAlloc(GMEM_MOVEABLE | GMEM_ZEROINIT, dwBitsSize);
	if (hDIB == NULL) NOT_ENOUGH_MEM ("Image");

	pHead = (LPBITMAPINFOHEADER) ::GlobalLock((HGLOBAL) hDIB);
	if (pHead == NULL) NOT_ENOUGH_MEM ("pHead");

	pHead->biSize    = sizeof (BITMAPINFOHEADER);
	pHead->biWidth   = width;   
	pHead->biHeight  = height; 
    pHead->biPlanes  = 1;  
	pHead->biBitCount = bPerPix * 8;  
	pHead->biCompression = 0; 
    pHead->biSizeImage = dwBitsPix; 
    pHead->biXPelsPerMeter = 0;   
	pHead->biYPelsPerMeter = 0; 
    pHead->biClrUsed = 0;    
	pHead->biClrImportant = 0; 

	bPerPixel = bPerPix;
	GetBitsPointer();
	Set2DVectors();

	::GlobalUnlock((HGLOBAL) hDIB);
}

CImage::~CImage()
{
	Free2DVectors();

	if (hDIB != NULL)
	{
		if (::GlobalFree((HGLOBAL) hDIB) != NULL) AfxMessageBox ("Error with global free !");
		hDIB = NULL;
	}

}

void CImage::Free2DVectors()
{
	if (Pixel != NULL) free (Pixel);	
}

BOOL CImage::ReadFromFile(CString filename)
{
	CFile file;
	CFileException fe;
	char buf[50];

	strcpy (buf,filename);
	if (!file.Open(buf, CFile::modeRead | CFile::shareDenyWrite, &fe))
	{
		
		AfxMessageBox ("Error while opening file");
		return FALSE;
	}

	// replace calls to Serialize with ReadDIBFile function
	TRY
	{
		hDIB = ::ReadDIBFile(file);
	}
	CATCH (CFileException, eLoad)
	{
		file.Abort(); // will not throw an exception
		AfxMessageBox ("Error while reading file");
		hDIB = NULL;
		return FALSE;
	}
	END_CATCH

	GetImageHeader();
	GetBitsPointer();
	Set2DVectors();
	
	// Compute the number of byte for one pixel
	bPerPixel = pHead->biBitCount / 8;

	return TRUE;
	
}

BOOL CImage::StoreInFile(CString filename)
{
	CFile file;
	CFileException fe;
	char buf[50];

	strcpy (buf,filename);

	if (!file.Open(filename, CFile::modeCreate |
	  CFile::modeReadWrite | CFile::shareExclusive, &fe))
	{
		AfxMessageBox ("Error while opening file");
		return FALSE;
	}

	// replace calls to Serialize with SaveDIB function
	BOOL bSuccess = FALSE;
	TRY
	{
		bSuccess = ::SaveDIB(hDIB, file);
		file.Close();
	}
	CATCH (CException, eSave)
	{
		file.Abort(); // will not throw an exception
		AfxMessageBox ("Error while writing file");
		return FALSE;
	}
	END_CATCH

	if (!bSuccess)
	{
		// may be other-style DIB (load supported but not save)
		//  or other problem in SaveDIB
		CString strMsg;
		strMsg = "Could not save DIB";
		MessageBox(NULL, strMsg, NULL, MB_ICONINFORMATION | MB_OK);
	}

	return bSuccess;
}

void CImage::DisplayImage(HDC pHdc,int xDest, int yDest, int dWidth, int dHeight,int fcol, int mode, int mask)
{
	LPSTR    lpDIBHdr;            // Pointer to BITMAPINFOHEADER
	lpDIBHdr  = (LPSTR) ::GlobalLock((HGLOBAL) hDIB);

	switch (mode)
	{
	case ALL_IMAGE	:
		
		SetDIBitsToDevice(pHdc,
						  xDest,
						  yDest,
						  pHead->biWidth,
						  pHead->biHeight,
						  0,0,0,
					      pHead->biHeight,
						  pBits,
						  (LPBITMAPINFO) lpDIBHdr,
						  DIB_RGB_COLORS
						 );
		break;

	case TRUNK_IMAGE	:

		StretchDIBits(pHdc,
					  xDest,
					  yDest,
					  dWidth,
					  dHeight,
					  fcol,
					  0,
					  dWidth,
					  pHead->biHeight,
					  pBits,
					  (LPBITMAPINFO) lpDIBHdr,
					  DIB_RGB_COLORS,
					  mask);
		break;

	case STRETCH_IMAGE :

		StretchDIBits(pHdc,
					  xDest,
					  yDest,
				      dWidth,
				      dHeight,
					  fcol,
					  0,
					  pHead->biWidth,
					  pHead->biHeight,
					  pBits,
					  (LPBITMAPINFO) lpDIBHdr,
					  DIB_RGB_COLORS,
					  SRCCOPY		  
					 );
		break;
	
	}
}

void CImage::Set2DVectors()
{
	int i;
	int		 SizeOfLine;				  // Size in bytes of a image line

	SizeOfLine = pHead->biBitCount / 8 * pHead->biWidth;

	Pixel = (BYTE **) malloc ( pHead->biHeight * sizeof(void *));
	Pixel[0] = (BYTE*) pBits;

	for (i = 1; i < pHead->biHeight; i++)
		Pixel[i] = (BYTE*) (pBits + i * SizeOfLine);
}

void CImage::GetImageHeader()
{
	LPSTR    lpDIBHdr;            // Pointer to BITMAPINFOHEADER

	lpDIBHdr  = (LPSTR) ::GlobalLock((HGLOBAL) hDIB);
	pHead = LPBITMAPINFOHEADER (lpDIBHdr);
}

void CImage::GetBitsPointer()
{
	LPSTR    lpDIBHdr;            // Pointer to BITMAPINFOHEADER

	lpDIBHdr   = (LPSTR) ::GlobalLock((HGLOBAL) hDIB);
	pBits	   = (BYTE*) ::FindDIBBits(lpDIBHdr);
}

void CImage::DisplayInfo()
{
	CString str = _T("");
	str.Format ("%sbiSize : %i\n",str,pHead->biSize);
	str.Format ("%sbiWidth : %i\n",str, pHead->biWidth);
	str.Format ("%sbiHeight : %i\n",str, pHead->biHeight);
	str.Format ("%sbiPlanes : %i\n",str, pHead->biPlanes);
	str.Format ("%sbiBitCount : %i\n",str,pHead->biBitCount);
	str.Format ("%sbiCompression : %i\n",str,pHead->biCompression);
	str.Format ("%sbiSizeImage : %i\n",str, pHead->biSizeImage);
	str.Format ("%sbiXPelsPerMeter : %i\n",str, pHead->biXPelsPerMeter);
	str.Format ("%sbiYPelsPerMeter : %i\n",str,pHead->biYPelsPerMeter);
	str.Format ("%sbibiClrUsed : %i\n", str,pHead->biClrUsed);
	str.Format ("%sbiClrImportant : %i\n",str, pHead->biClrImportant);
	AfxMessageBox (str);
}

void CImage::GetPixelRGB(int nCol, int nLin, BYTE *r, BYTE *g, BYTE *b)
{
	int adrOffset; 	
	adrOffset = nCol * bPerPixel;
	
	*b = Pixel[nLin][adrOffset++];
	*g = Pixel[nLin][adrOffset++];
	*r = Pixel[nLin][adrOffset];
}

void CImage::GetPixelHSI(int nCol, int nLin, BYTE *h, BYTE *s, BYTE *i)
{
	int adrOffset; 	
	adrOffset = nCol * bPerPixel;
	
	*h = Pixel[nLin][adrOffset++];
	*s = Pixel[nLin][adrOffset++];
	*i = Pixel[nLin][adrOffset];
}

void CImage::SetPixel(int nCol, int nLin, BYTE r, BYTE g, BYTE b)
{
	int adrOffset; 	
	adrOffset = nCol * bPerPixel;
	
	Pixel[nLin][adrOffset++] = b;
	Pixel[nLin][adrOffset++] = g;
	Pixel[nLin][adrOffset]	 = r;	
}

// Copy pixels between sfrom sto (source image) to pixels from dfrom (destination img)
void CImage::CopyToImage(CImage *dest,int dfrom, int sfrom, int sto, short mode)
{
	int i,j;
	GetBitsPointer();
	switch (mode)
	{
		case ALL_IMAGE :

			if (dest->pHead->biSizeImage != pHead->biSizeImage)
			{
				AfxMessageBox ("The size of the two images are different !");
				return;
			}

			for (i = 0; i < (int) pHead->biSizeImage; i++)
				dest->pBits[i] = pBits[i];
		break;

		case TRUNK_IMAGE :
			
			dfrom = dfrom * bPerPixel;
			sfrom = sfrom * bPerPixel;
			sto   = sto   * bPerPixel;
			for (i = 0; i < pHead->biHeight; i++)
			{
				for (j = sfrom; j < sto; j++)			
					dest->Pixel[i][j+dfrom-sfrom] = Pixel[i][j];
			}
		break;

	}
}

void CImage::Filter_3x3(CImage *dest, short chan, BOOL fast, MATRIX_3X3 mat)
{
	int lig, col, index, height, width, t1, t2, t3;
	int wTmp;
	WORD *pW;

	height = pHead->biHeight-1;
	width  = pHead->biWidth-1;


	switch (chan)
	{
		case (RG_CHANNEL)  :
		case (RGB_CHANNEL) :
		{
			if (fast)
			{
				for (lig = 1; lig < height; lig++)
				for (col = 1; col < width; col++)
				{		
					t1 = (col-1)*bPerPixel;
					t2 = (col+1)*bPerPixel;

					pW =  (WORD *) &Pixel[lig-1][t1];
					wTmp  =  - (*pW);

					pW =  (WORD *) &Pixel[lig-1][t2];
					wTmp +=  *pW;

					pW =  (WORD *) &Pixel[lig  ][t1];
					wTmp +=  *pW * -2;

					pW =  (WORD *) &Pixel[lig  ][t2];
					wTmp +=  *pW * 2;

					pW =  (WORD *) &Pixel[lig+1][t1];
					wTmp +=  -(*pW);

					pW =  (WORD *) &Pixel[lig+1][t2];
					wTmp +=  *pW;

					// Take the absolute value of the gradient
					wTmp = abs (wTmp); 

				}
			}
			else
			{
				for (lig = 1; lig < height; lig++)
				for (col = 1; col < width; col++)
				{		
					index = col * bPerPixel;
					t1 = (col-1)*bPerPixel;
					t2 = (col+1)*bPerPixel;

					pW =  (WORD *) &Pixel[lig-1][t1];
					wTmp  =  *pW * mat[0][0];

					pW =  (WORD *) &Pixel[lig-1][index];
					wTmp +=  *pW * mat[0][1];

					pW =  (WORD *) &Pixel[lig-1][t2];
					wTmp +=  *pW * mat[0][2];

					pW =  (WORD *) &Pixel[lig  ][t1];
					wTmp +=  *pW * mat[1][0];

					pW =  (WORD *) &Pixel[lig  ][index];
					wTmp +=  *pW * mat[1][1];

					pW =  (WORD *) &Pixel[lig  ][t2];
					wTmp +=  *pW * mat[1][2];

					pW =  (WORD *) &Pixel[lig+1][t1];
					wTmp +=  *pW * mat[2][0];
	
					pW =  (WORD *) &Pixel[lig+1][index];
					wTmp +=  *pW * mat[2][1];

					pW =  (WORD *) &Pixel[lig+1][t2];
					wTmp +=  *pW * mat[2][2];

					// Take the absolute value of the gradient
					wTmp = abs (wTmp); 
		
					// Update image for display
					dest->Pixel[lig][index++] = LOBYTE (wTmp);
					dest->Pixel[lig][index++] = LOBYTE (wTmp);
					dest->Pixel[lig][ index ] = LOBYTE (wTmp);

				}
			}
			break;
		}

		default :
		{
			if (fast)
			{
				for (lig = 1; lig < height; lig++)
				for (col = 1; col < width; col++)
				{
					t1 = (col-1)*bPerPixel+chan;
					t2 = (col+1)*bPerPixel+chan;
					wTmp  = -Pixel[lig-1][t1] 
						  +  Pixel[lig-1][t2] 
				  		  -2 * Pixel[lig  ][t1] 
						  +2 * Pixel[lig  ][t2] 
				 		  -  Pixel[lig+1][t1] 
						  +  Pixel[lig+1][t2];

					// Take the absolute value of the gradient
					wTmp = abs (wTmp); 
				}
			}
			else
			{	
				for (lig = 1; lig < height; lig++)
				for (col = 1; col < width; col++)
				{
					index = col * bPerPixel;
					t1 = (col-1)*bPerPixel+chan;
					t2 = (col+1)*bPerPixel+chan;
					t3 = index+chan;
					wTmp  = Pixel[lig-1][t1] * mat[0][0]
						  + Pixel[lig-1][t3] * mat[0][1]
						  + Pixel[lig-1][t2] * mat[0][2]
				  		  + Pixel[lig  ][t1] * mat[1][0]
						  + Pixel[lig  ][t3] * mat[1][1]
						  + Pixel[lig  ][t2] * mat[1][2]
				 		  + Pixel[lig+1][t1] * mat[2][0]
						  + Pixel[lig+1][t3] * mat[2][1]
						  + Pixel[lig+1][t2] * mat[2][2];

					// Take the absolute value of the gradient
					wTmp = abs (wTmp); 
		
					dest->Pixel[lig][index++] = LOBYTE (wTmp);
					dest->Pixel[lig][index++] = LOBYTE (wTmp);
					dest->Pixel[lig][ index ] = LOBYTE (wTmp);
				}
			}
			break;
		}
	}
}

void CImage::ExtractChannel(CImage *dest, short channel)
{
	int i,val;
	int index, PixSize;
	BYTE  b;
	WORD  *pw;

	PixSize = pHead->biWidth * pHead->biHeight;

	switch (channel)
	{
		case RED_CHANNEL :	
			for (i = 0; i < PixSize; i++)
			{
				index = i*bPerPixel;
				b = pBits[index+2];
				dest->pBits[index]   = b;
				dest->pBits[index+1] = b;
				dest->pBits[index+2] = b;
			}
			break;

		case GREEN_CHANNEL :
			for (i = 0; i < PixSize; i++)
			{
				index = i*bPerPixel;
				b = pBits[index+1];
				dest->pBits[index]   = b;
				dest->pBits[index+1] = b;
				dest->pBits[index+2] = b;
			}
			break;

		case BLUE_CHANNEL :
			for (i = 0; i < PixSize; i++)
			{
				index = i*bPerPixel;
				b = pBits[index];
				dest->pBits[index]   = b;
				dest->pBits[index+1] = b;
				dest->pBits[index+2] = b;
			}
			break;

		case GRAY_CHANNEL :
			for (i = 0; i < PixSize; i++)
			{
				index = i*bPerPixel;
				val = pBits[index]
					+ pBits[index+1] 
				    + pBits[index+2];
				b = (BYTE) (val/3);
				dest->pBits[index]   = b;
				dest->pBits[index+1] = b;
				dest->pBits[index+2] = b;
			}
			break;

		case (RG_CHANNEL):
			for (i = 0; i < PixSize; i++)
			{
				// Store the sum in the two low bytes
				index = i*bPerPixel;
				pw = (WORD*) (dest->pBits + index);
				*pw = pBits[index+1] + pBits[index+2];
			}
			break;

		case (RGB_CHANNEL) :
			for (i = 0; i < PixSize; i++)
			{
				// Store the sum in the two low bytes
				index = i*bPerPixel;
				pw = (WORD*) (dest->pBits + index);
				*pw = pBits[index] + pBits[index+1] + pBits[index+2];
			}
			break;

		default : AfxMessageBox ("This channel doesn't exists!"); break;
	}
}

void CImage::HSI_Values (CImage *dest)
{
        int i;
        int r, g, b,max, min, hue, saturation, temp;  // long !!!
		register int max_min, intensity; 

        for (i=0; i < (int) pHead->biSizeImage; i += bPerPixel)
        {
                b = pBits[i];
                g = pBits[i+1];
                r = pBits[i+2];

                max = MAX3(r,g,b);           /* compute Intensity */
                min = MIN3(r,g,b);
				max_min = max - min;
                intensity = (min + max) / 2;
                dest->pBits[i+2] = (unsigned char) intensity; // intensity in 3th byte 

                if (max == min)                         /* achromatic case */
                {
                        dest->pBits[i] = 0;      // hue is undefined
                        dest->pBits[i+1] = 0;    // saturation = 0
						continue;
                }
                else if (intensity < 20)        /* too dark (ADDED) */
                {
                        dest->pBits[i] = 0;      // hue is undefined
                        dest->pBits[i+1] = 0;    // saturation = 0
						continue;
                }
                else if (intensity > 235)       /* too bright (ADDED) */
                {
                        dest->pBits[i] = 0;      // hue is undefined
                        dest->pBits[i+1] = 0;    // saturation = 0
						continue;
                }
                else            /* chromatic case */
                {
                        if (intensity <= 127)   /* bottom cone (dark) */
                                saturation = (255 * (max_min)) / (max + min);
                        else    /* top cone (bright) */
                                saturation = (255 * (max_min)) / (512 - max - min);

                        dest->pBits[i+1] = (unsigned char) saturation; // saturation in 2th byte 

                        if (r == max)
                                hue = (60 *(g -b))/(max_min);
        /* color is between yellow and magenta */
                        else if (g == max)
                                hue = 60*2 + (60*(b-r))/(max_min);
        /* color is between cyan and yellow */
                        else
                                hue = 60*4 + (60*(r-g))/(max_min);      
        /* color is between magenta and cyan */
                        if (hue < 0) hue = hue + 360;
                        
                        temp = (255 * hue) / 360;
                        dest->pBits[i] = (unsigned char) temp; // hue in 1th byte 
                }
        }
}

void CImage::AfxHSI(int nCol, int nLine)
{
	CString str;
	BYTE H,S,I;
	GetPixelHSI (nCol, nLine, &H, &S, &I);
	str.Format ("Pixel x = %i y = %i\n\nH : %u\nS : %u\nI  : %u",nCol, nLine, H, S, I);
	AfxMessageBox (str);
}
	
void CImage::AfxRGB(int nCol, int nLine)
{
	CString str;
	BYTE R,G,B;
	GetPixelRGB (nCol, nLine, &R, &G, &B);
	str.Format ("Pixel x = %i y = %i\n\nR : %u\nG : %u\nB : %u",nCol, nLine, R, G, B);
	AfxMessageBox (str);
}

void CImage::TresholdImage(CImage *dest, short chan, short treshlow, short treshhigh, int *HueHist, short mode)
{	
	register BYTE val;
	register short thigh = treshhigh,tlow = treshlow;
	register int idx, idxchan, PixelSize;
	
	register int i;
	
	// Because all bytes have the same value
	if (chan == GRAY_CHANNEL) chan = GREEN_CHANNEL;

	PixelSize = pHead->biWidth * pHead->biHeight;

	if (mode == MODE_NORMAL)
	{
		for (i = 0; i < PixelSize; i++)
		{
			idx		= i * bPerPixel;
			idxchan = idx + chan;
		
			if ( (pBits[idxchan] >= tlow) && (pBits[idxchan] <= thigh) )
			{
				val = pBits[idxchan];
				dest->pBits[idx++]   = val;
				dest->pBits[idx++]   = val;
				dest->pBits[idx]     = val;
				continue;
			}
			else
			{
				dest->pBits[idx++]   = 0;
				dest->pBits[idx++]   = 0;
				dest->pBits[idx]     = 0;
			}
		}
	}
	else
	if (mode == MODE_LET_PIXEL_INFO_IN)
	{
		for (i = 0; i < PixelSize; i++)
		{
			idx		= i * bPerPixel;
			idxchan = idx + chan;
		
			if ( (pBits[idxchan] >= tlow) && (pBits[idxchan] <= thigh) )
			{
				dest->pBits[idx]   = pBits[idx];
				dest->pBits[idx+1] = pBits[idx+1];
				dest->pBits[idx+2] = pBits[idx+2];
				continue;
			}
			else
			{
				dest->pBits[idx++]   = 0;
				dest->pBits[idx++]   = 0;
				dest->pBits[idx]     = 0;
			}
		}
	}
	else
	if (mode == MODE_LET_PIXEL_INFO_IN_AND_HISTOGRAM)
	{
		// Purge histogram
		for (i = 0; i < MAX_HUE+1; i++) HueHist[i] = 0;

		for (i = 0; i < PixelSize; i++)
		{
			idx		= i * bPerPixel;
			idxchan = idx + chan;
		
			if ( (pBits[idxchan] >= tlow) && (pBits[idxchan] <= thigh) )
			{
				dest->pBits[idx]   = pBits[idx];
				dest->pBits[idx+1] = pBits[idx+1];
				dest->pBits[idx+2] = pBits[idx+2];
				HueHist[pBits[idx+HUE_CHANNEL]] ++;
				continue;
			}
			else
			{
				dest->pBits[idx++]   = 0;
				dest->pBits[idx++]   = 0;
				dest->pBits[idx]     = 0;
			}
		}
	}
	else
	if (mode == MODE_PASS_ZERO)
	{
		for (i = 0; i < PixelSize; i++)
		{
			idx		= i * bPerPixel;
			idxchan = idx + chan;
		
			if ( (pBits[idxchan] >= tlow) || (pBits[idxchan] <= thigh) )
			{
				val = pBits[idxchan];
				dest->pBits[idx++]  = val;
				dest->pBits[idx++]  = val;
				dest->pBits[idx]    = val;
				continue;
			}
			else
			{	
				dest->pBits[idx++]  = 0; 
				dest->pBits[idx++]  = 0;
				dest->pBits[idx]	= 0;
			}
		}
	}	
}

void CImage::ComputeColHistogram(int *hist, short middle, short range, short addmax)
{
	int		 maxi = -1;
	register int lig, col, idx;
	register short bPerPix = bPerPixel; 	
	register short dif;
	register float frange = float (range);
	register float slope = float (- addmax / range);

	for (int i = 0; i < pHead->biWidth; i++) hist[i] = 0;
	
	for (lig = 0; lig < pHead->biHeight; lig++)
		for (col = 0; col < pHead->biWidth; col++)
		{		
			idx = col*bPerPix;
			// Add only if saturation != 0
			if (Pixel[lig][idx+SAT_CHANNEL] != 0) 
			{
				dif = abs (middle - Pixel[lig][idx+HUE_CHANNEL]);
				if (dif > range) dif = 256 - dif;
				hist[col] += int (slope * dif + addmax);
			}	
		}
}
